unit Main;

interface

uses
  Windows, SysUtils, Messages, Classes, Forms, ExtCtrls, Controls,
  StdCtrls, Math, Dialogs, ComCtrls, StrUtils, ShellAPI, IniFiles, Registry,
  Graphics, Menus, LychrelThread;

type
  TMainForm = class(TForm)
    Label9: TLabel;
    TrayIcon1: TTrayIcon;
    PopupMenu1: TPopupMenu;
    Beenden1: TMenuItem;
    Fensterffnen1: TMenuItem;
    GroupBox1: TGroupBox;
    Label10: TLabel;
    Label14: TLabel;
    Label15: TLabel;
    Label16: TLabel;
    GroupBox2: TGroupBox;
    CheckBox1: TCheckBox;
    Label11: TLabel;
    GroupBox3: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label12: TLabel;
    Label13: TLabel;
    Label17: TLabel;
    Edit1: TComboBox;
    Label18: TLabel;
    Label19: TLabel;
    Button3: TButton;
    Button4: TButton;
    Button2: TButton;
    Label5: TLabel;
    Edit2: TEdit;
    Button1: TButton;
    Button5: TButton;
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure TrayIcon1DblClick(Sender: TObject);
    procedure OnMinimize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Fensterffnen1Click(Sender: TObject);
    procedure LinkClick(Sender: TObject);
    procedure Edit1KeyPress(Sender: TObject; var Key: Char);
    procedure CheckBox1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Beenden1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    conf_saveinterval: integer;
    KeineExitNachfrage: boolean;
    Ly: TLychrelSolve;
    procedure KonfigurationSpeichern();
    procedure LychrelStatus(digitcount, iterations, speed: string);
    procedure Solved(initialzahl, digits, iterationcounter: string);
    procedure WMQueryEndSession(var Msg: TWMQueryEndSession); message WM_QueryEndSession;
  end;

var
  MainForm: TMainForm;

implementation

uses
  UBigIntsV3, ISF_Utilities;

{$R *.dfm}

procedure TMainForm.Beenden1Click(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.Button2Click(Sender: TObject);
begin
  ShellExecute(Application.Handle, 'open', 'Data\', '', '', SW_NORMAL);
end;

procedure TMainForm.CheckBox1Click(Sender: TObject);
var
  reg: tregistry;
const
  ValName = 'VTS-Lychrel-Solver';
begin
  reg := tregistry.create;
  try
    reg.RootKey := HKEY_LOCAL_MACHINE;
    if reg.OpenKey('Software\Microsoft\Windows\CurrentVersion\Run', False) then
    begin
      if mainform.CheckBox1.Checked then
      begin
        reg.WriteString(ValName, '"'+ParamStr(0)+'" -silent');
      end
      else
      begin
        reg.DeleteValue(ValName)
      end;
      reg.CloseKey;
    end
    else
    begin
      reg.RootKey := HKEY_CURRENT_USER;
      if reg.OpenKey('Software\Microsoft\Windows\CurrentVersion\Run', False) then
      begin
        if mainform.CheckBox1.Checked then
        begin
          reg.WriteString(ValName, '"'+ParamStr(0)+'" -silent');
        end
        else
        begin
          reg.DeleteValue(ValName)
        end;
        reg.CloseKey;
      end
      else
      begin
        CheckBox1.Checked := not CheckBox1.Checked;
      end;
    end;
  finally
    reg.free;
  end;
end;

procedure TMainForm.Edit1KeyPress(Sender: TObject; var Key: Char);
begin
  // #8 is Backspace
  if not (Key in [#8, '0'..'9']) then
  begin
    Key := #0;
  end;
end;

procedure TMainForm.Fensterffnen1Click(Sender: TObject);
begin
  TrayIcon1DblClick(self);
end;

procedure TMainForm.KonfigurationSpeichern();
var
  IniFile: TMemIniFile;
begin
  IniFile := TMemIniFile.Create(ExtractFilePath(ParamStr(0))+
                                'Data\Settings.ini');
  try
    if MainForm.Edit1.Text <> '' then
    begin
      IniFile.WriteString('Settings', 'CurrentSearch', MainForm.Edit1.Text);
    end;
    if (MainForm.Edit2.Text <> '') and (MainForm.Edit2.Text <> '0') then
    begin
      IniFile.WriteString('Settings', 'AutoSaveInterval', MainForm.Edit2.Text);
    end;
    IniFile.UpdateFile;
  finally
    IniFile.free;
  end;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Hide;
  trayicon1.Visible := false;

  KonfigurationSpeichern;

  if assigned(Ly) then
  begin
    Ly.Terminate;
    Ly.Free;
  end;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if KeineExitNachfrage then
  begin
    CanClose := true;
  end
  else
  begin
    CanClose := MessageDlg('Shutdown the application?', mtConfirmation, [mbYes,mbNo], 0) = mrYes;
  end;
end;

procedure TMainForm.LychrelStatus(digitcount, iterations, speed: string);
begin
  label4.caption := digitcount;
  label2.caption := iterations;
  label13.caption := speed;
end;

function add_separators(input: string): string;
var
  i1: tinteger;
begin
  i1 := TInteger.create;
  try
    i1.Assign(input);
    result := i1.ConvertToDecimalString(true);
  finally
    i1.Free;
  end;
end;

procedure TMainForm.Solved(initialzahl, digits, iterationcounter: string);
begin
  Label12.Caption    := 'Solved!';
  Label12.Font.Color := clGreen;
  Label13.Visible    := False;

  MessageDlg('Initial value ' + initialzahl +
             ' became a palindrome with ' + digits +
             ' digits after ' + iterationcounter +
             ' iterations!' + #13#10#13#10 +
             'Please send the result files to info@daniel-marschall.de !',
             mtInformation, [mbOK], 0);
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  IniFile: TMemIniFile;
  reg: TRegistry;
  a, b, check_initial_value: string;
  isf_status: T_ISF_Status;
  cas_status: T_CAS_STATUS;
  ainitialzahl, aaktuell, aiterationcounter, acas: string;
begin
  Caption := Label9.Caption;
  application.OnMinimize := OnMinimize;
  //TrayIcon1.Visible := not Application.ShowMainForm;

  if not directoryexists(ExtractFilePath(ParamStr(0))+'Data\') then
  begin
    mkdir(ExtractFilePath(ParamStr(0))+'Data\');
  end;

  IniFile := TMemIniFile.Create(ExtractFilePath(ParamStr(0))+
                                'Data\Settings.ini');
  try
    ainitialzahl := IniFile.ReadString('Settings', 'CurrentSearch', label8.Caption);
    if strtoint(ainitialzahl) < 1 then
    begin
      MessageDlg('Initial value may not be smaller than 1! The program will terminate.', mtError, [mbOK], 0);
      KeineExitNachfrage := true;
      Close;
      exit;
    end;
    conf_saveinterval := IniFile.ReadInteger('Settings', 'AutoSaveInterval', strtoint(edit2.text));
  finally
    IniFile.free;
  end;

  edit2.text        := inttostr(conf_saveinterval);
  edit1.Text        := ainitialzahl;
  label8.Caption    := FloatToStrF(strtoint(ainitialzahl), ffNumber, 10, 0);
  trayicon1.hint    := Label9.Caption + ' ['+label8.Caption+']';

  isf_status := load_isf(ExtractFilePath(ParamStr(0))+'Data\'+ainitialzahl+'_master.isf', check_initial_value, aiterationcounter, aaktuell);

  if isf_status = isf_not_exists then
  begin
    aiterationcounter := '0';
    aaktuell := ainitialzahl;
  end
  else
  begin
    if isf_status <> isf_ok then
    begin
      if isf_status = isf_mod9_error then
        MessageDlg('The file Data\'+ainitialzahl+'_master.isf did not pass the MOD-9 check. The program will terminate.', mtError, [mbOK], 0);

      if isf_status = isf_crc16_error then
        MessageDlg('The file Data\'+ainitialzahl+'_master.isf has a bad CRC-16 checksum. The program will terminate.', mtError, [mbOK], 0);

      if isf_status = isf_digit_error then
        MessageDlg('The file Data\'+ainitialzahl+'_master.isf has inconsistent digit numbers. The program will terminate.', mtError, [mbOK], 0);

      KeineExitNachfrage := true;
      close;
      exit;
    end;

    if check_initial_value <> ainitialzahl then
    begin
      MessageDlg('The file Data\'+ainitialzahl+'_master.isf doesn''t contain the expected initial value. The program will terminate.', mtError, [mbOK], 0);
      KeineExitNachfrage := true;
      close;
      exit;
    end;
  end;

  // Cascade-Information laden
  // Fraglich: CAS-Funktionalitt optinal lassen? Bei CAS-Fehlern Weiterarbeit ermglichen?

  cas_status := cas_load(ExtractFilePath(ParamStr(0))+'Data\'+ainitialzahl+'_master.cas', aCas);

  if cas_status = cas_not_exists then
  begin
    if isf_status <> isf_not_exists then
    begin
      MessageDlg('The file Data\'+ainitialzahl+'_master.cas does not exists. The program will terminate.', mtError, [mbOK], 0);
      KeineExitNachfrage := true;
      close;
      exit;
    end;
  end
  else
  begin
    if cas_status <> cas_ok then
    begin
      if cas_status = cas_crc16_error then
        MessageDlg('The file Data\'+ainitialzahl+'_master.cas has a bad CRC-16 checksum. The program will terminate.', mtError, [mbOK], 0);

      if cas_status = cas_invalid_character then
        MessageDlg('The file Data\'+ainitialzahl+'_master.cas contains invalid characters. The program will terminate.', mtError, [mbOK], 0);

      if cas_status = cas_digit_error then
        MessageDlg('The file Data\'+ainitialzahl+'_master.cas has a wrong amount of digits. The program will terminate.', mtError, [mbOK], 0);

      if cas_status = cas_initialvalue_error then
        MessageDlg('The file Data\'+ainitialzahl+'_master.cas has a bad checksum of the initial value. The program will terminate.', mtError, [mbOK], 0);

      KeineExitNachfrage := true;
      close;
      exit;
    end;
  end;

  reg := tregistry.create;
  try
    reg.RootKey := HKEY_LOCAL_MACHINE;
    reg.OpenKey('Software\Microsoft\Windows\CurrentVersion\Run', False);
    a := reg.ReadString('VTS-Lychrel-Solver');
    reg.CloseKey;

    reg.RootKey := HKEY_CURRENT_USER;
    reg.OpenKey('Software\Microsoft\Windows\CurrentVersion\Run', False);
    b := reg.ReadString('VTS-Lychrel-Solver');
    reg.CloseKey;

    CheckBox1.Checked := (LowerCase(a) = '"'+LowerCase(ParamStr(0)+'" -silent')) or
                         (LowerCase(b) = '"'+LowerCase(ParamStr(0)+'" -silent'));
  finally
    reg.free;
  end;

  // Prfen, ob Herunterfahren-Verhalten in Registry OK

  reg := tregistry.Create;
  try
    reg.RootKey := HKEY_CURRENT_USER;
    reg.OpenKey('Control Panel\Desktop', False);
    if reg.ReadString('AutoEndTasks') <> '0' then
    begin
      reg.WriteString('AutoEndTasks', '0');
      MessageDlg('The program has updated your shutdown configuration.'+#13#10+
                 '"AutoEndTasks" from "HKCU\Control Panel\Desktop" turned into 0.', mtInformation, [mbOK], 0);
    end;
    reg.CloseKey;
  finally
    reg.free;
  end;

  // Lychrel-Prozess starten

  Ly := TLychrelSolve.create(ainitialzahl, aaktuell, aiterationcounter, acas, conf_saveinterval);
  Ly.Priority := tpLowest; //tpIdle;
  Ly.SolvedCallback := Solved;
  Ly.Resume;
end;

procedure TMainForm.FormHide(Sender: TObject);
begin
  Ly.StatusCallback := nil;
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  Ly.StatusCallback := LychrelStatus;
end;

procedure TMainForm.LinkClick(Sender: TObject);
begin
  shellexecute(handle, 'open', pchar('http://'+tlabel(sender).caption+'/'), '', '', SW_SHOW);
end;

procedure TMainForm.OnMinimize(Sender: TObject);
begin
  //TrayIcon1.Visible := True;
  ShowWindow(Application.Handle, SW_HIDE);
end;

procedure TMainForm.TrayIcon1DblClick(Sender: TObject);
begin
  ShowWindow(Application.Handle, SW_SHOW);
  MainForm.Show;
  //TrayIcon1.Visible := False;

  application.Restore;
  application.BringToFront;
end;

procedure TMainForm.WMQueryEndSession(var Msg: TWMQueryEndSession);
begin
  KeineExitNachfrage := true;
  Msg.Result := 1;
end;

end.
